PROCEDURE Com_XTab  && Compute CrossTab Baded on Parameters
PARAMETERS p_topfld, p_sidfld, p_calfld, p_calc
*------------------------------------------------------------------
* Compute data normalized form of a relational database 
* file to a spreadsheet-like structure.  It summarizes the 
* data in one field by expressing it in terms of two other 
* fields.
* 
* SYNTAX 
* 
* DO Com_XTab WITH <top field name>, <side field name>, ;
*                  <calc field>, <type calc>
* 
* where <type calc> can be one of the following functions:
* 
*   CNT
*   MAX
*   MIN
*   SUM
* 
* USAGE
* 
* The Com_XTab command handles all records in the active 
* file that match any filter condition placed on the data.
* Com_XTab places the result of the CROSSTAB in a file
* named CROSSTAB.  Com_XTab will also place the column names
* for the <top field name> in a file named COLUMNS.  
* Com_XTab will overwrite these files.
* 
* OPTIONS
* 
* CNT counts the records that match the <top field name>
* and the <side field name>.
* 
* SUM adds together the values of <calc field> that match
* the <top field name> and the <side field name>.
* 
* MAX determines the maximum value of <calc field> for
* matching values of <top field name> and <side field name>.
* 
* MIN determines the minimum value of <calc field> for
* matching values of <top field name> and <side field name>.
* 
* EXAMPLE
* 
* Consider the following database file.
* 
* CAMPING   Name  Item     Price
* -------   ----  -----   ------
*     1     John  Tent    245.00
*     2     Mark  Stove    59.00
*     3     Jane  Heater   75.00
*     4     Mark  Stove    79.00
*     5     Jane  Heater   95.00
*     6     Jane  Tent    325.00
*     7     John  Tent    175.00
*     8     Mark  Stove    89.00
* 
* DO Com_XTab WITH "Item", "Name", "Price", "SUM"
* 
*   will create the following two files:
* 
* CROSSTAB  Name   COL01   COL02   COL03
* --------  ----  ------  ------  ------
*     1     Jane  170.00    0.00  325.00  
*     2     John    0.00    0.00  420.00
*     3     Mark    0.00  227.00    0.00
* 
* The CROSSTAB file will be active in work area 1.
* 
* COLUMNS   Column  TopField
* -------   ------  --------
*     1     COL01   Heater
*     2     COL02   Stove
*     3     COL03   Tent
* 
* The COLUMNS file will be open in work area 2.  The COLUMNS
* file will have an MDX TAG Column on the Column field.  
* 
* COM_XTAB will create a SET FIELDS TO command for each column
* based on the Columns file to get the correct column heading.
* This technique will only work if the resulting column name
* is less than or equal to 10 characters and it contains
* no punctuation or spaces.
* 
* You may use this file in the report designer to place a 
* meaningfull column label for a report.  Before entering
* the report designer execute the command:
*     SET FIELDS TO <Enter>
* 
* SEE ALSO
* 
*------------------------------------------------------------------

*-- Save the current environment
  SET FILTER TO                 && Clear the filter condition if any
  PRIVATE lc_talk
  IF SET("TALK") = "ON"         && Set talk off, but restore it later
    SET TALK OFF
    lc_talk = "ON"
  ELSE
    lc_talk = "OFF"
  ENDIF

  PRIVATE lc_safety
  lc_safety = SET("SAFETY")     && Save the SAFETY state
  SET SAFETY OFF                && and turn it off.

  PRIVATE lc_excl
  lc_excl = SET("EXCLUSIVE")    && Save the EXCLUSIVE state
  SET EXCLUSIVE ON              && and turn it on.

  PRIVATE lc_status
  lc_status = SET("STATUS")     && Save the status bar before turning it off
  SET STATUS OFF

  SET CURSOR OFF
  lc_stat1 = "Determining the number of UNIQUE Records"
  lc_stat2 = "Creating CROSSTAB and COLUMNS DBF files "
  lc_stat3 = "Populating the COLUMNS file             "
  lc_stat4 = "Populating the CROSSTAB file            "
  lc_stat5 = "Creating Pseudo Field Headings          "
  DEFINE WINDOW com_xtab FROM 5,9 TO 13, 70
  DO ShadowG WITH 5, 9, 13, 70
  ACTIVATE WINDOW com_xtab

  @ 0, 20 SAY "Computing CrossTab"
  @ 2, 5 SAY lc_stat1 + + CHR(205) + CHR(205) + CHR(16)
  @ 2, 50 SAY "Working"

*-- Make sure a data file is open
  IF ISBLANK( ALIAS() )
    lc_error = XT_ErrMsg( "No file is in use!", "" )
    DO XT_RestEv
    CANCEL
  ENDIF   

*-- Preserve the order of the file going into crosstab
  lc_order = ORDER()

*-- Strip the alias from the field id it exists
  ln_apos   = AT( "->", p_topfld )  && Alias position, optional       
  IF ln_apos > 0                    && Alias exists, now get the field
    lc_ptopfld = SUBSTR( p_topfld, ln_apos + 2, LEN( p_topfld ) - ln_apos )
  ELSE
    lc_ptopfld = UPPER( p_topfld )
  ENDIF

  ln_apos   = AT( "->", p_sidfld )  && Alias position, optional       
  IF ln_apos > 0                    && Alias exists, now get the field
    lc_psidfld = SUBSTR( p_sidfld, ln_apos + 2, LEN( p_sidfld ) - ln_apos )
  ELSE
    lc_psidfld = UPPER( p_sidfld )
  ENDIF

  ln_apos   = AT( "->", p_calfld )  && Alias position, optional       
  IF ln_apos > 0                    && Alias exists, now get the field
    lc_pcalfld = SUBSTR( p_calfld, ln_apos + 2, LEN( p_calfld ) - ln_apos )
  ELSE
    lc_pcalfld = UPPER( p_calfld )
  ENDIF


*-- Verify the calculation type
  IF .NOT. p_calc $ "SUM,CNT,MAX,MIN,AVG"
    lc_error = XT_ErrMsg( "Calculation type: " + p_calc + ;
                         " is not supported!", "" )
    DO XT_RestEv
    CANCEL
  ENDIF

*-- Compute the number of top and side values
  lc_typetop = TYPE( lc_ptopfld )
  DO CASE
    CASE lc_typetop = "L"
      INDEX ON IIF( &lc_ptopfld., "T", "F" ) TAG topfld UNIQUE
    CASE lc_typetop = "C"
      INDEX ON UPPER( &lc_ptopfld. ) TAG topfld UNIQUE
    CASE lc_typetop $ "NF"
      INDEX ON &lc_ptopfld. TAG topfld UNIQUE
    CASE lc_typetop = "D"
      INDEX ON DTOS( &lc_ptopfld. ) TAG topfld UNIQUE
  ENDCASE

  COUNT TO ln_columns
  IF ln_columns > 254
    lc_error = XT_ErrMsg( "Unique columns for the top exceeds 254", "")
    DELETE TAG topfld
    DO XT_RestEv
    CANCEL
  ENDIF

  lc_typesid = TYPE( lc_psidfld )
  DO CASE
    CASE lc_typesid = "L"
      INDEX ON IIF( &lc_psidfld., "T", "F" ) TAG sidfld UNIQUE
    CASE lc_typesid = "C"
      INDEX ON UPPER( &lc_psidfld. ) TAG sidfld UNIQUE
    CASE lc_typesid $ "NF"
      INDEX ON &lc_psidfld. TAG sidfld UNIQUE
    CASE lc_typesid = "D"
      INDEX ON DTOS( &lc_psidfld. ) TAG sidfld UNIQUE
  ENDCASE

  COUNT TO ln_rows

  @ 2, 50 SAY "Complete"
  @ 3, 5 SAY lc_stat2 + + CHR(205) + CHR(205) + CHR(16)
  @ 3, 50 SAY "Working"

*-- Create the CROSSTAB file to hold the values
  COPY TO _cross STRUCTURE EXTENDED
  SELECT SELECT()
  USE _cross        

*-- Make the calc field's attributes global for use in the SET FIELDS TO
*-- command after the crosstab is created to adjust the column size.
  IF TYPE( "gn_dec" ) <> "U"
    RELEASE gn_dec
  ENDIF
  PUBLIC gn_dec
  IF TYPE( "gn_len" ) <> "U"
    RELEASE gn_len
  ENDIF
  PUBLIC gn_len

*-- Get the attributes for the side field
  LOCATE FOR field_name = UPPER( lc_psidfld )
  ln_lensid = field_len
  ln_decsid = field_dec

*-- Get the attributes for the top field
  LOCATE FOR field_name = UPPER( lc_ptopfld )
  ln_lentop = field_len
  ln_dectop = field_dec

*-- Get the attributes for the calculated column
  LOCATE FOR field_name = UPPER( lc_pcalfld )
  IF p_calc = "CNT"
    gn_len = LEN( ALLTRIM( RECCOUNT() ) )
    gn_dec = 0
  ELSE
    gn_len = field_len
    IF p_calc = "SUM" .AND. gn_len < 16   && Adjust column length if SUM
      gn_len = gn_len + 2                 && calculation is desired
    ENDIF
    gn_dec = field_dec
  ENDIF
  ZAP

*----------------------------------------------------------------------
*-- Create the CROSSTAB file fields
*----------------------------------------------------------------------
*-- Make the Side Column column
  APPEND BLANK
  REPLACE field_name WITH lc_psidfld, ;
          field_type WITH IIF( lc_typesid = "L", "C", lc_typesid ), ;
          field_len  WITH ln_lensid , ;
          field_dec  WITH ln_decsid , ;
          field_idx  WITH "N"

*-- Make the crosstab columns
  ln_col = 1                      && Counter for creating columns
  DO WHILE ln_col <= ln_columns
    lc_colname = "COL" + SUBSTR( STR( 100 + ln_col, 3, 0), 2, 2 )
    APPEND BLANK
    REPLACE field_name WITH lc_colname, ;
            field_type WITH "N"       , ;
            field_len  WITH gn_len, ;
            field_dec  WITH gn_dec    , ;
            field_idx  WITH "N"

    ln_col = ln_col + 1           && Increment column counter
  ENDDO

  CREATE crosstab FROM _cross
  USE crosstab ALIAS crosstab 
  DO CASE
    CASE lc_typesid = "L"
      INDEX ON IIF( &lc_psidfld., "T", "F" ) TAG &lc_psidfld
    CASE lc_typesid = "C"
      INDEX ON UPPER( &lc_psidfld. ) TAG &lc_psidfld
    CASE lc_typesid $ "NF"
      INDEX ON &lc_psidfld. TAG &lc_psidfld
    CASE lc_typesid = "D"
      INDEX ON DTOS( &lc_psidfld. ) TAG &lc_psidfld
  ENDCASE

*----------------------------------------------------------------------
*-- Make the COLUMN file fields
*----------------------------------------------------------------------
  SELECT SELECT()
  USE _cross
  ZAP
  APPEND BLANK
  REPLACE field_name WITH "COLUMN", ;
          field_type WITH "C"     , ;
          field_len  WITH 5       , ;
          field_dec  WITH 0       , ;
          field_idx  WITH "Y"
  APPEND BLANK            
  REPLACE field_name WITH lc_ptopfld, ;
          field_type WITH "C"       , ;
          field_len  WITH ln_lentop , ;
          field_dec  WITH 0         , ;
          field_idx  WITH "N"

  CREATE columns FROM _cross
  USE columns ALIAS columns
  DO CASE
    CASE lc_typetop = "L"
      INDEX ON IIF( &lc_ptopfld., "T", "F" ) TAG &lc_ptopfld
    CASE lc_typetop = "C"
      INDEX ON UPPER( &lc_ptopfld. ) TAG &lc_ptopfld
    CASE lc_typetop $ "NF"
      INDEX ON &lc_ptopfld. TAG &lc_ptopfld
    CASE lc_typetop = "D"
      INDEX ON DTOS( &lc_ptopfld. ) TAG &lc_ptopfld
  ENDCASE

  @ 3, 50 SAY "Complete"
  @ 4, 5 SAY lc_stat3 + + CHR(205) + CHR(205) + CHR(16)
  @ 4, 50 SAY "Working"

*-- Fill in the column file with the UNIQUE values for the top fields
*-- from the source DBF
  SELECT 1
  SET ORDER TO TAG topfld
  ln_col = 1
  SCAN      
    lc_colname =  "COL" + SUBSTR( STR( 100 + ln_col, 3, 0), 2, 2 )
    SELECT columns
    APPEND BLANK
    DO CASE 
      CASE lc_typetop = "C"
        REPLACE column      WITH lc_colname, ;
                &lc_ptopfld WITH &p_topfld
      CASE lc_typetop = "L"
        REPLACE column      WITH lc_colname, ;
                &lc_ptopfld WITH IIF( &p_topfld., "T", "F" )

      CASE lc_typetop $ "NF"
        REPLACE column      WITH lc_colname, ;
                &lc_ptopfld WITH ;
                  ALLTRIM( STR( &p_topfld, ln_lentop, ln_dectop ) )
      CASE lc_typetop = "D"
        REPLACE column      WITH lc_colname, ;
                &lc_ptopfld WITH DTOC( &p_topfld. )
    ENDCASE

    ln_col = ln_col + 1
    SELECT 1
  ENDSCAN

*-- Initialize the crosstab matrix by filling in the side column
*-- with the UNIQUE values from the source DBF
*-- This two pass approach is no faster than filling it in as
*-- the crosstab function is populated.  It has the added advantage
*-- of correctly ordering the side column.
  SELECT 1
  SET ORDER TO TAG sidfld
  SCAN
    SELECT crosstab
    APPEND BLANK
    REPLACE &lc_psidfld WITH ;
            IIF(lc_typesid = "L", ;
                IIF( &p_sidfld, "T", "F" ),;
                &p_sidfld ;
            )
    SELECT 1
  ENDSCAN

  @ 4, 50 SAY "Complete"
  @ 5, 5 SAY lc_stat4 + + CHR(205) + CHR(205) + CHR(16)
  @ 5, 50 SAY "Working"

*-- Compute the matrix - brute force
  SELECT 1
  SET ORDER TO &lc_order          && Return to the original order
*                                 && in the event a FOR condition was
*                                 && placed on the index.
  SET FILTER TO &gc_filter
  SCAN
    IF lc_typesid = "L"
      l_seekval = IIF( &lc_psidfld., "T", "F" )
    ELSE
      IF lc_typesid = "C"
        l_seekval = UPPER( &lc_psidfld )
      ELSE
        l_seekval = &lc_psidfld
      ENDIF
    ENDIF

    IF SEEK( l_seekval, "crosstab" )

      DO CASE
        CASE lc_typetop = "C"
          l_seekval = UPPER( &lc_ptopfld )
        CASE lc_typetop = "L"
          l_seekval = IIF( &lc_ptopfld., "T", "F" )
        CASE lc_typetop $ "NF"
          l_seekval = ALLTRIM( STR( &lc_ptopfld, ln_lentop, ln_dectop ) )
        CASE lc_typetop = "D"
          l_seekval = DTOC( &lc_ptopfld )
      ENDCASE

      IF SEEK( l_seekval, "columns" )
        lc_colname = columns->column
        DO CASE
          CASE p_calc = "SUM"
            REPLACE crosstab->&lc_colname WITH ;
                    &p_calfld + crosstab->&lc_colname
          CASE p_calc = "CNT"
            REPLACE crosstab->&lc_colname WITH ;
                    1 + crosstab->&lc_colname
          CASE p_calc = "MAX"
            REPLACE crosstab->&lc_colname WITH ;
                    MAX( &p_calfld, crosstab->&lc_colname )
          CASE p_calc = "MIN"
            IF crosstab->&lc_colname = 0
              REPLACE crosstab->&lc_colname WITH &p_calfld
            ELSE
              REPLACE crosstab->&lc_colname WITH ;
                      MIN( &p_calfld, crosstab->&lc_colname )
            ENDIF
        ENDCASE
      ENDIF

    ENDIF

  ENDSCAN

  @ 5, 50 SAY "Complete"
  @ 6, 5 SAY lc_stat5 + + CHR(205) + CHR(205) + CHR(16)
  @ 6, 50 SAY "Working"

*-- Remove the UNIQUE Tags on the data file
  DELETE TAG topfld
  DELETE TAG sidfld

*-- Put the CROSSTAB and columns into view
  CLOSE DATABASE
  SELECT 1
  USE crosstab
  USE columns IN 2 ORDER TAG column

*-- Create the field column labels
  SET FIELDS TO &lc_psidfld. /R
  SET FIELDS OFF
  ln_col = 1
  DO WHILE ln_col <= ln_columns
    lc_colname = "COL" + SUBSTR( STR( 100 + ln_col, 3, 0), 2, 2 )
    lc_coltitl = LOOKUP( columns->&lc_ptopfld, lc_colname, columns->column )
    lc_coltitl = XT_ClnTl( (lc_coltitl) )
    SET FIELDS TO &lc_coltitl = ( &lc_colname )
    SET FIELDS OFF
    ln_col = ln_col + 1
  ENDDO
  SET FIELDS ON

  @ 6, 50 SAY "Complete"
  DO XT_RestEv
RETURN
*-- EOP: Com_XTab


PROCEDURE XT_RestEv && UDF - Com_XTab
*------------------------------------------------------------------
* Restore the previous environment
*------------------------------------------------------------------
  SET TALK &lc_talk
  SET SAFETY &lc_safety
  SET EXCLUSIVE &lc_excl
  SET STATUS &lc_status
  SET CURSOR ON
  RELEASE WINDOW com_xtab
  RELEASE gn_len, gn_dec, lc_cursor, lc_status, lc_escape, lc_proc
RETURN
*-- EOP: XT_RestEv


FUNCTION XT_IsFld   && UDS - Com_XTab
PARAMETER p_fldname
*----------------------------------------------------------------------------
* Synopsis:
* 
* XT_isfld( <p_fldname> )
* 
* The IF_FIELD() function strips off the alais, if any, from the field
* <p_fldname> and then calls IsInView() to verify that the field is in the
* current view.
*
*----------------------------------------------------------------------------
  PRIVATE ll_result, ln_fldcnt

  lc_pfield = UPPER( p_fldname )    && Force the field name to upper case
  ln_apos   = AT( "->", p_fldname ) && Alias position, optional
  IF ln_apos > 0                    && Alias exists, now get it
    lc_alchk = SUBSTR( p_fldname, 1, ln_apos - 1 )
    lc_pfield = SUBSTR( p_fldname, ln_apos + 2, LEN( p_fldname ) - ln_apos )
  ENDIF
  lc_exact = SET("EXACT")
  SET EXACT OFF

  ll_result = IsInView( lc_pfield ) && Check to see if field is in view

  SET EXACT &lc_exact

RETURN( ll_result )
*-- EOF: XT_isfld( <p_filename> )


FUNCTION XT_ErrMsg  && UDS - Com_XTab
PARAMETER p_msg, p_pads
*-----------------------------------------------------------------------
* Synopsis:
*
* XT_ErrMsg( <p_msg>, <p_pads> )
*
* The XT_ErrMsg() function will display the message <p_msg> inside of
* a "radio" dialog box.  The user can select on of the comma delimited
* options specified by the expression <p_pads>.  The XT_ErrMsg() function
* will return the option selected by the user.
*
*-----------------------------------------------------------------------
PRIVATE lc_select     && The pad name the user selected

  DEFINE WINDOW XT_ErrMsg FROM 5, 9 TO 11, 71 DOUBLE
  ACTIVATE WINDOW XT_ErrMsg
  IF ISBLANK( p_pads )
    ? p_msg
    ?
    ? "Press any key to continue"
    ln_key = INKEY( 0 )
    lc_select = "CANCEL"
  ELSE
    ? p_msg                        && Stub code
    ?
    ? "Press any key to continue"  && Stub code
    ln_key = INKEY( 0 )            && Stub code
    lc_select = "CANCEL"
  ENDIF
  DEACTIVATE WINDOW XT_ErrMsg
  RELEASE WINDOW XT_ErrMsg 

RETURN( lc_select )
*-- EOF: XT_ErrMsg( <p_msg>, <p_pads> )


FUNCTION XT_ClnTl   && UDS - Com_XTab
PARAMETER p_title
*-----------------------------------------------------------------------
*
* Synopsis:
* 
* XT_ClnTl( <p_title> )
*
* The XT_ClnTl() function will change any characters in the top row
* title column that are not valid field name characters to the 
* underscore "_" character.  If the value for <p_title> begins with
* a numeral, the XT_ClnTl() function will add the "F" character to
* the beginning of <p_title>.  If the value for <p_title> begins 
* with a character other than a numeric or letter, XT_ClnTl() will
* replace it with the "F" character.
*
* Example:
*
* XT_ClnTl("91.92") will equal "F91_92"
* XT_ClnTl("AT&T")  will equal "AT_T"
*
*-----------------------------------------------------------------------
  PRIVATE lc_result, ln_lentitl, ln_cnt, lc_valchars, lc_char

  lc_valchars = "ABCDEFGHIJKLMNOPQRSTUVWXYZ_0123456789"
  lc_val1st = "ABCDEFGHIJKLMNOPQRSTUVWXYZ"

  ln_lentitl = LEN( RTRIM( p_title) )  && get the length of the column
  IF ln_lentitl > 10                   && set the length to 10 if greater
    ln_lentitl = 10                    && that the number of chars in a field
  ENDIF

  ln_cnt = 1
  lc_result = ""

*-- Check for a valid first character, if it is not, use a "F" for
*-- field instead.
  lc_char = UPPER( SUBSTR( p_title, ln_cnt, 1 ) )  
  IF .NOT. lc_char $ lc_val1st                     
    p_title = "F" + p_title
    ln_lentitl = ln_lentitl + 1
  ENDIF

*-- Loop through the characters and make the substitution is needed
  DO WHILE ln_cnt <= ln_lentitl
    lc_char = UPPER( SUBSTR( p_title, ln_cnt, 1 ) )
    IF  lc_char $ lc_valchars
      lc_result = lc_result + lc_char
    ELSE
      lc_result = lc_result + "_"
    ENDIF
    ln_cnt = ln_cnt + 1
  ENDDO

RETURN( lc_result )
*-- EOF: XT_ClnTl( <p_title> )


